/////////// Network.cpp /////////////////////////////////////////////////////////
// Network support implementation file
//
// Version:			1.00
// Author:			Robert Minnis
// Dependencies:	Network.h			All platforms
//					ProjNet.h			All platforms
//					Except.h			All platforms
//					Messages.h			All platforms
//					stdafx.h			Windows Only
//					TRIForm.h			Windows Only
//					ResDefs.h			Windows Only
//					DPLobby.h			Windows Only
//					CGUID.h				Windows Only
// History:			01/10/97 - version 0.10 released
//					01/20/97 - version 0.20 released
//					11/03/97 - version 1.00 released
//
////////////////////////////////////////////////////////////////////////////////

#include "Network.h"
#include "ProjNet.h"
#include "TRIStr.h"
#include "TRIList.h"
#include "TRIThrow.h"

#ifdef OLDMSGSYSTEM
	#include "messages.h"
#else
	#include "NewMsg.h"
#endif

#include <string.h>
#include <stdlib.h>

////////////////////////////////////////////////////////////////
// Uncomment the following line to log network activity
//#define NETWORK_LOG

#ifdef NETWORK_LOG
	static short gLogIndent = 0;
	void LogIndent(short inIndent)
	{
		gLogIndent += inIndent;
	}

	void Log(char *inString)
	{
		static FILE *file = NULL;

		if(file == NULL)
			file = fopen("network.log", "a+t");

		if(file != NULL)
		{
			for(short i=0; i<gLogIndent; i++)
				fputs("\t", file);

			if(inString != NULL)
				fputs(inString, file);

			fputs("\n", file);

			fclose(file);
			file = NULL;
		}
	}

	void LogEx(char *inString, short inLine)
	{
		static long space = 80;
		long len = strlen(inString);
		char msg[1024];
		memset(msg, ' ', sizeof(msg));

		strcpy(msg, inString);
		msg[len] = ' ';

		char line[25];
		sprintf(line, "(line: %d)", inLine);

//		if(len > space)
//			space = len + 2;

		strcpy(msg+space, line);

		Log(msg);
	}

	#ifdef _DEBUG
		#define LOGINDENT LogIndent
		#define LOG(inString) LogEx(inString, __LINE__)
	#else
		#define LOGINDENT LogIndent
		#define LOG Log
	#endif
#else
	#define LOGINDENT(indent)
	#define LOG(str)
#endif

////////////////////////////////////////////////////////////////
// constant definitions
const short		kPingTimeout = 5;		// wait 5 seconds for a synchronous ping to complete
const unsigned long	kTimeToSeconds = 65536*18;	// const to convert from getTime to seconds
const SPlayer	*kAllPlayers = NULL;	// Send to all players
const unsigned char kBuiltInType = 0x80; // Marks internal packet types


////////////////////////////////////////////////////////////////
// network initialization & termination routines
CTRINetwork gNoNetwork;
CTRINetwork *gNetwork = &gNoNetwork;	// A network object must always exist

CNetworkManager gNetworkManager;

void CreateDefaultNetwork(void)
{
	LOG("");
	LOG("*** Starting network ************************************************");

	// Platform dependent protocol selection
	#ifdef NETWORK_PLAY
		#ifdef WIN
			typedef CTRIDPlay CNetwork;
		#else
			#ifdef MAC
				typedef CNetSprocket CNetwork;
			#else
				typedef CTRINetwork CNetwork;
			#endif
		#endif
	#else
		typedef CTRINetwork CNetwork;
	#endif

	SAutoConn *autoconn = gNetwork->RemoveAutoConnInfo();
	DestroyNetwork();
	gNetwork = new CNetwork;
	if(gNetwork == NULL)
		gNetwork = &gNoNetwork;
	else
		gNetwork->InstallAutoConnInfo(autoconn);
}

void DestroyNetwork(void)
{
	if(gNetwork != &gNoNetwork)
	{
		delete gNetwork;
		gNetwork = &gNoNetwork;

		LOG("*** Network closed");
	}
}

SPacket *MakeVariablePacket(PacketID inPacketType, unsigned long inSize, EByteOrder inByteOrder)
{
	unsigned long totalSize = inSize + sizeof(SPacket);
	SPacket *packet = (SPacket*) new char[totalSize];

	packet->mLocalData.mSize = totalSize;
	packet->mLocalData.mFrom = NULL;
	packet->mLocalData.mOwner = SPacket::eOther;

	packet->mID = inPacketType;
	packet->mByteOrder = inByteOrder;
	packet->mFromID = -1;

	return packet;
}

void DeleteVariablePacket(SPacket *inPacket)
{
	delete inPacket;
}


const unsigned long kDefaultType = 0;

////////////////////////////////////////////////////////////////
// Methods for generic classes
SProtocol::SProtocol(const String& inName, unsigned long inType)
	: SChangeable(inType)
{
	mName = new String(inName);
}

SProtocol::~SProtocol(void)
{
	delete mName;
}

SSession::SSession(const String& inName, const String& inPassword, short inMaxPlayers, unsigned long inType)
	: SChangeable(inType)
{
	mName = new String(inName);
	mPassword = new String(inPassword);
	mMaxPlayers=inMaxPlayers;
}

SSession::~SSession(void)
{
	delete mName;
	delete mPassword;
}

Boolean
SSession::NeedsPassword(void)
{
	return mPassword->Length() > 0;
}

SAutoConn::SAutoConn(unsigned long inType)
	: SChangeable(inType)
{
	mAddress = NULL;
}

SAutoConn::~SAutoConn(void)
{
	delete mAddress;
}

SPlayer::SPlayer(const String& inName, PlayerID inID, unsigned long inType, EOwner inOwner)
	: SChangeable(inType)
{
	mName = new String(inName);
	mID=inID;
	mPingTime=-1;
	mNotPresent=kFalse;
	mNotActive=kFalse;
	mLocal=kFalse;
	mHost=kFalse;
	mStatus=0;
	mOwner=inOwner;
}

SPlayer::~SPlayer(void)
{
	delete mName;
}

SPlayer&
SPlayer::operator=(const SPlayer& inPlayer)
{
	delete mName;
	mName = new String(*inPlayer.mName);
	mID = inPlayer.mID;
	mPingTime = inPlayer.mPingTime;
	mNotPresent = inPlayer.mNotPresent;
	mNotActive = inPlayer.mNotActive;
	mLocal = inPlayer.mLocal;
	mHost = inPlayer.mHost;
	mStatus = inPlayer.mStatus;
	mOwner = inPlayer.mOwner;

	return *this;
}


////////////////////////////////////////////////////////////////
// Private packet types handled by CNetworkManager & CTRINetwork
STARTPACKET(PacketPlayerAvailable, kBuiltInType+1)
	char		mName[80];
	PlayerID	mID;
ENDPACKET

STARTPACKET(PacketPlayerActive, kBuiltInType+2)
	char		mName[80];
	PlayerID	mID;
	Boolean		mNotifyOnly;
ENDPACKET

STARTPACKET(PacketPlayerRemove, kBuiltInType+3)
	PlayerID	mID;
ENDPACKET

STARTPACKET(PacketPing, kBuiltInType+4)
	unsigned long mSendTime;
ENDPACKET

STARTPACKET(PacketPingResponse, kBuiltInType+5)
	unsigned long mSendTime;
ENDPACKET

STARTPACKET(PacketNameChanged, kBuiltInType+6)
	PlayerID	mID;
	char		mNewName[80];
ENDPACKET

STARTPACKET(PacketFinalize, kBuiltInType+7)
	PlayerID		mID;
	ElementIndex	mIndex;
ENDPACKET

STARTPACKET(PacketChangeHost, kBuiltInType+8)
	PlayerID		mID;
ENDPACKET

//Defined in header
//STARTPACKET(PacketRadioData, 0x89)
//	// This is a variable length packet used by the radio class
//ENDPACKET

STARTPACKET(PacketPlayerListStatus, kBuiltInType+10)
	Boolean	mClosed;
	Boolean	mAllowReplacements;
	Boolean	mAllowAdditions;
ENDPACKET


////////////////////////////////////////////////////////////////
// CNetworkManager
CNetworkManager::CNetworkManager(void)
{
}

CNetworkManager::~CNetworkManager(void)
{
}

SPacket*
CNetworkManager::MakePacket(PacketID inID, SPlayer *inPlayerRef, SPacket *inPacketRef, void *inOther)
{
	SPacket *packet = NULL;
	switch(inID)
	{
		case PacketChat::eID:
			{
			PacketChat *apacket = new PacketChat(kNativeByteOrder, SPacket::eNetMgr);
			strcpy(apacket->mMessage, (char*) inOther);
			packet = apacket;
			}
			break;
		case PacketRadioData::eID:
			{
			PacketRadioData *apacket = new PacketRadioData(kNativeByteOrder, SPacket::eNetMgr);
			packet = apacket;
			}
			break;
		case PacketPlayerAvailable::eID:
			{
			PacketPlayerAvailable *apacket = new PacketPlayerAvailable(kNativeByteOrder, SPacket::eNetMgr);
			strcpy(apacket->mName, *inPlayerRef->mName);
			apacket->mID = inPlayerRef->mID;
			packet = apacket;
			}
			break;
		case PacketPlayerActive::eID:
			{
			PacketPlayerActive *apacket = new PacketPlayerActive(kNativeByteOrder, SPacket::eNetMgr);
			strcpy(apacket->mName, *inPlayerRef->mName);
			apacket->mID = inPlayerRef->mID;
			apacket->mNotifyOnly = kFalse;
			packet = apacket;
			}
			break;
		case PacketPlayerRemove::eID:
			{
			PacketPlayerRemove *apacket = new PacketPlayerRemove(kNativeByteOrder, SPacket::eNetMgr);
			apacket->mID = inPlayerRef->mID;
			packet = apacket;
			}
			break;
		case PacketPing::eID:
			{
			PacketPing *apacket = new PacketPing(kNativeByteOrder, SPacket::eNetMgr);
			apacket->mSendTime = getTime();
			packet = apacket;
			}
			break;
		case PacketPingResponse::eID:
			{
			PacketPingResponse *apacket = new PacketPingResponse(kNativeByteOrder, SPacket::eNetMgr);
			apacket->mSendTime = ((PacketPing*) inPacketRef)->mSendTime;
			packet = apacket;
			}
			break;
		case PacketNameChanged::eID:
			{
			PacketNameChanged *apacket = new PacketNameChanged(kNativeByteOrder, SPacket::eNetMgr);
			apacket->mID = inPlayerRef->mID;
			strcpy(apacket->mNewName, *inPlayerRef->mName);
			packet = apacket;
			}
			break;
		case PacketFinalize::eID:
			{
			PacketFinalize *apacket = new PacketFinalize(kNativeByteOrder, SPacket::eNetMgr);
			apacket->mID = inPlayerRef->mID;
			apacket->mIndex = *((ElementIndex*)inOther);
			packet = apacket;
			}
			break;

		case PacketPlayerListStatus::eID:
			{
			PacketPlayerListStatus *apacket = new PacketPlayerListStatus(kNativeByteOrder, SPacket::eNetMgr);
			apacket->mClosed = gNetwork->IsPlayerListClosed();
			apacket->mAllowReplacements = gNetwork->AllowReplacements();
			apacket->mAllowAdditions = gNetwork->AllowAdditions();
			packet = apacket;
			}
			break;

		case PacketChangeHost::eID:
			{
			PacketChangeHost *apacket = new PacketChangeHost(kNativeByteOrder, SPacket::eNetMgr);
			apacket->mID = inPlayerRef->mID;
			packet = apacket;
			}
			break;
	}

	return packet;
}

void
CNetworkManager::DeletePacket(SPacket *inPacket)
{
	if(inPacket->mLocalData.mOwner == SPacket::eNetMgr)
		delete inPacket;
}

void
CNetworkManager::ProcessPacket(SPacket *inPacket)
{
	switch(inPacket->mID)
	{
		case PacketChat::eID:
			{
			PacketChat *packet = (PacketChat*) inPacket;
			SendMsg(Msg_Chat, Group_Network, packet);
			}
			break;
		case PacketRadioData::eID:
			{
			PacketRadioData *packet = (PacketRadioData*) inPacket;
			SendMsg(Msg_ReceivedRadio, Group_Network, packet);
			}
			break;
		default:
			break;
	}
}

SPlayer*
CNetworkManager::CreatePlayer(const String& inName, PlayerID inID, unsigned long inType, SPlayer::EOwner inOwner)
{
	return new SPlayer(inName, inID, inType, inOwner);
}

////////////////////////////////////////////////////////////////
// network initialization & termination routines
CTRINetwork::CTRINetwork(void)
{
	mStatus = eNoNetSupport;
	mAutoConn = NULL;
	mAllowHostMigration = kTrue;
	mDeliverLocalPackets = kFalse;
	mNetworkManager = NULL;
	mLocalPlayerID = 0;
	mLocalPlayerIndex = 1;	// Set to one by default for backwards compatibility
	mBestPacketRate = 8;	// Default to 8 packets per second

	mCurProtocol = NULL;
	mCurSession = NULL;

	mPlayersClosed = kFalse;
	mAllowReplacements = kFalse;
	mAllowAdditions = kFalse;

	mWaitingForPacket = kFalse;
	mSyncWaitPacketType = 0;

	mAvailablePlayers = new SPlayerList;
	if(mAvailablePlayers == NULL) TRIThrow("Cannot create available player list", ERR_InternalError);

	mActivePlayers = new SPlayerList;
	if(mActivePlayers == NULL) TRIThrow("Cannot create active player list", ERR_InternalError);

	RegisterNetworkManager(&gNetworkManager);
}

CTRINetwork::~CTRINetwork(void)
{
	Disconnect();

	delete mCurProtocol;
	delete mCurSession;

	delete mAvailablePlayers;
	delete mActivePlayers;
	delete mAutoConn;
}

void
CTRINetwork::StoreAutoConnectInfo(const String& inAddress, Boolean inHost,
		unsigned long inProtocol)
{
	mAutoConn = new SAutoConn;
	if(mAutoConn != NULL)
	{
		mAutoConn->mProtocolSelector = inProtocol;
		mAutoConn->mAddress = new String(inAddress);
		mAutoConn->mHost = inHost;
	}
}

Boolean
CTRINetwork::StartAutoConnect(const String& inPlayerName)
{
	if(HasConnection() || !CanAutoConnect()) return kFalse;

	SPlayer *player = CreateLocalPlayer(inPlayerName);
	if(StartAutoConnect(player))
		return kTrue;
	else
	{
		delete player;
		return kFalse;
	}
}

Boolean
CTRINetwork::StartAutoConnect(SPlayer *inPlayer)
{
	if(HasConnection() || !CanAutoConnect()) return kFalse;

	LOGINDENT(1);
	LOG("CTRINetwork::StartAutoConnect");

	SProtocol *protocol = NULL;
	SSession *session = NULL;

	LOG(" - Get Protocol");
	protocol = GetAutoConnectProtocol(-1);
	if(protocol == NULL) goto finish; //return kFalse;

	LOG(" - Get Session");
	session = GetAutoConnectSession(protocol, mAutoConn->mHost);
	if(session == NULL) goto finish; //return kFalse;

	LOG(" - Start");
	if(mAutoConn->mHost)
		OpenHostSession(protocol, session);
	else
		OpenJoinSession(protocol, session);

	if(HasConnection())
	{
		LOG(" - Has Connection");
		UseLocalPlayer(inPlayer);
	}
	SetStatus(eConnected);
	SendMsg(Msg_Connect, Group_Network);

finish:
	LOG(" - Complete");
	LOGINDENT(-1);

	delete mAutoConn;
	mAutoConn = NULL;

	return HasConnection();
}

SSession*
CTRINetwork::CreateSessionObject(const String& inName, const String& inPassword, short inMaxPlayers)
{
	return new SSession(inName, inPassword, inMaxPlayers);
}

void
CTRINetwork::ClearProtocolList(SProtocolList& inList)
{
	SProtocolIter iter(&inList);
	while(!iter.AtEnd())
	{
		SProtocol *protocol = iter.Next();
		if(protocol != GetCurrentProtocol())
			delete protocol;
	}

	inList.RemoveAll();
}

void
CTRINetwork::ClearSessionList(SSessionList& inList)
{
	SSessionIter iter(&inList);
	while(!iter.AtEnd())
	{
		SSession *session = (SSession*) iter.Next();
		if(session != GetCurrentSession())
			delete session;
	}

	inList.RemoveAll();
}

Boolean
CTRINetwork::Disconnect(void)
{
	if(IsDisconnecting()) return kFalse;

	if(HasConnection())
	{
		LOGINDENT(1);
		LOG("CTRINetwork::Disconnect");

		mPlayersClosed = kFalse;	// Force the list to be opened so that the players are removed

		SetStatus(eDisconnecting);

		RemoveAllPlayers();

		SetStatus(eNotConnected);
		SetCurrentProtocol(NULL);
		SetCurrentSession(NULL);

		SendMsg(Msg_Disconnect, Group_Network);

		LOGINDENT(-1);
		return kTrue;
	}
	return kFalse;
}

SPlayer*
CTRINetwork::CreateLocalPlayer(const String& inName)
{
	SPlayer *player = mNetworkManager->CreatePlayer(inName, -1, kDefaultType, SPlayer::eNetwork);
	player->mHost = IsHosting();

	UseLocalPlayer(player);
	return player;
}

Boolean
CTRINetwork::UseLocalPlayer(SPlayer *inPlayer)
{
	if(!HasConnection()) return kFalse;
	if(inPlayer == NULL) return kFalse;

	SPlayer *oldPlayer = GetLocalPlayer();
	if(oldPlayer != NULL)
	{	// Local player is being changed
		// Subclasses must support creating the first local player
		Boolean	replaced = kFalse;
		Boolean	nameChanged = (oldPlayer!=inPlayer) && (oldPlayer->mName != inPlayer->mName);

		if(oldPlayer != inPlayer)
		{
			// Replace player entry
			if(!replaced)
			{
				SPlayerIter iter(mAvailablePlayers);
				while(!iter.AtEnd())
				{
					SPlayer *p = iter.Next();
					if(p == oldPlayer)
					{
						(*mAvailablePlayers)[iter.CurrentIndex()] = inPlayer;
						replaced = kTrue;
						break;
					}
				}
			}

			if(!replaced)
			{
				SPlayerIter iter(mActivePlayers);
				while(!iter.AtEnd())
				{
					SPlayer *p = iter.Next();
					if(p == oldPlayer)
					{
						(*mActivePlayers)[iter.CurrentIndex()] = inPlayer;
						replaced = kTrue;
						break;
					}
				}
			}

			if(replaced)
			{
				// Copy over pertinent player information
				inPlayer->mID = oldPlayer->mID;
				inPlayer->mPingTime = oldPlayer->mPingTime;
				inPlayer->mNotPresent = oldPlayer->mNotPresent;
				inPlayer->mNotActive = oldPlayer->mNotActive;
				inPlayer->mLocal = oldPlayer->mLocal;
				inPlayer->mHost = oldPlayer->mHost;
				inPlayer->mStatus = oldPlayer->mStatus;

				// delete oldPlayer
				if(oldPlayer->mOwner == SPlayer::eNetwork)
					delete oldPlayer;
			}
		}
		
		if(replaced || nameChanged)
		{
			ChangedLocalPlayer();
		}

		return kTrue;
	}

	return kFalse;
}

void
CTRINetwork::ChangedLocalPlayer(void)
{
	if(!HasConnection()) return;

	SPlayer *player = GetLocalPlayer();
	if(player != NULL)
	{
		SendMsg(Msg_LocalPlayerChanged, Group_Network, player);
		SendMsg(Msg_PlayerChanged, Group_Network, player);
		SendPacket(PacketNameChanged::eID, kAllPlayers, player);
	}
}

SPlayer*
CTRINetwork::GetLocalPlayer(void)
{
	SPlayerIter iter(mAvailablePlayers);
	while(!iter.AtEnd())
	{
		SPlayer *player = iter.Next();
		if(player->mLocal)
			return player;
	}

	iter.SetList(mActivePlayers);
	while(!iter.AtEnd())
	{
		SPlayer *player = iter.Next();
		if(player->mLocal)
			return player;
	}

	return NULL;
}

SPlayer*
CTRINetwork::GetHost(void)
{
	SPlayerIter iter(mAvailablePlayers);
	while(!iter.AtEnd())
	{
		SPlayer *player = iter.Next();
		if(player->mHost)
			return player;
	}

	iter.SetList(mActivePlayers);
	while(!iter.AtEnd())
	{
		SPlayer *player = iter.Next();
		if(player->mHost)
			return player;
	}

	return NULL;
}

SPlayer*
CTRINetwork::GetPlayer(PlayerID inPlayerID)
{
	SPlayerIter iter(mActivePlayers);
	while(!iter.AtEnd())
	{
		SPlayer *player = iter.Next();
		if(player->mID == inPlayerID)
			return player;
	}

	iter.SetList(mAvailablePlayers);
	while(!iter.AtEnd())
	{
		SPlayer *player = iter.Next();
		if(player->mID == inPlayerID)
			return player;
	}

	return NULL;
}

SPlayer*
CTRINetwork::GetPlayer(PlayerIndex inPlayerIndex)
{
	if(inPlayerIndex < 0)
		return NULL;

#ifdef _DEBUG
	if(inPlayerIndex > mActivePlayers->GetCount()-1)
		TRIThrow(String("Invalid player index requested (index=<<1>>)") % (long) inPlayerIndex, ERR_InternalError);
#endif

	return (*mActivePlayers)[inPlayerIndex+1];
}

PlayerIndex
CTRINetwork::GetPlayerIndex(SPlayer *inPlayer)
{
	SPlayerIter iter(GetActivePlayers());
	while(!iter.AtEnd())
	{
		if(iter.Next() == inPlayer)
		{
			return iter.CurrentIndex() - 1;
		}
	}

	return -1;
}

Boolean
CTRINetwork::MakePlayerActive(SPlayer *inPlayer, Boolean inNotifyOnly)
{
	if(!HasConnection()) return kFalse;

	if(IsPlayerActive(inPlayer)) return kTrue;

	Boolean added = kFalse;

	if(!added && mPlayersClosed && mAllowReplacements)
	{
		if(IsPlayerAvailable(inPlayer))
		{
			SPlayerIter iter(mActivePlayers);
			while(!iter.AtEnd())
			{
				SPlayer *player = iter.Next();
				if(player->mNotPresent)
				{
					RemoveAvailablePlayer(inPlayer);
					SPlayer *oldPlayer = (*mActivePlayers)[iter.CurrentIndex()];
					if(oldPlayer->mOwner == SPlayer::eNetwork)
						delete oldPlayer;
					(*mActivePlayers)[iter.CurrentIndex()] = inPlayer;
					added = kTrue;
					break;
				}
			}
		}
	}

	if(!added && ((!mPlayersClosed) || (mPlayersClosed && mAllowAdditions)))
	{
		if(RemoveAvailablePlayer(inPlayer))
		{
			mActivePlayers->Insert(inPlayer);
			added = kTrue;
		}
	}

	if(added)
	{
		if(IsHosting())
		{
			SendPacket(PacketPlayerActive::eID, kAllPlayers, inPlayer);
		}

		if(inNotifyOnly)
			SendMsg(Msg_NotifyActivePlayer, Group_Network, inPlayer);
		else
			SendMsg(Msg_AddActivePlayer, Group_Network, inPlayer);
	}

	return added;
}

Boolean
CTRINetwork::MakePlayerInactive(SPlayer *inPlayer)
{
	if(!HasConnection()) return kFalse;

	if(mPlayersClosed) return kFalse;

	if(IsPlayerAvailable(inPlayer)) return kTrue;

	if(RemoveActivePlayer(inPlayer))
	{
		mAvailablePlayers->Insert(inPlayer);
		SendMsg(Msg_AddAvailPlayer, Group_Network, inPlayer);

		if(IsHosting())
		{
			SendPacket(PacketPlayerAvailable::eID, kAllPlayers, inPlayer);
		}

		return kTrue;
	}

	return kFalse;
}

Boolean
CTRINetwork::RemovePlayer(SPlayer *inPlayer)
{
	if(!HasConnection()) return kFalse;

	if(RemoveAvailablePlayer(inPlayer))
	{
		if(IsHosting())
		{
			SendPacket(PacketPlayerRemove::eID, kAllPlayers, inPlayer);
		}

		SendMsg(Msg_RemoveAvailPlayer, Group_Network, inPlayer);
		if(inPlayer->mOwner == SPlayer::eNetwork)
			delete inPlayer;

		return kTrue;
	}

	if(RemoveActivePlayer(inPlayer))
	{
		if(IsHosting() && !IsDisconnecting())
		{
			SendPacket(PacketPlayerRemove::eID, kAllPlayers, inPlayer);
		}

		if((mPlayersClosed && inPlayer->mNotPresent) || !mPlayersClosed)
			SendMsg(Msg_RemoveActivePlayer, Group_Network, inPlayer);

		if(!mPlayersClosed && inPlayer->mOwner==SPlayer::eNetwork)
			delete inPlayer;

		return kTrue;
	}

	return kFalse;
}

Boolean
CTRINetwork::RemoveAvailablePlayer(SPlayer *inPlayer)
{
	SPlayerIter iter(mAvailablePlayers);
	while(!iter.AtEnd())
	{
		SPlayer *player = iter.Next();
		if(player == inPlayer)
		{
			mAvailablePlayers->Remove(iter.CurrentIndex());
			return kTrue;
		}
	}

	return kFalse;
}

Boolean
CTRINetwork::RemoveActivePlayer(SPlayer *inPlayer)
{
	SPlayerIter iter(mActivePlayers);
	while(!iter.AtEnd())
	{
		SPlayer *player = iter.Next();
		if(player == inPlayer)
		{
			if(mPlayersClosed)
			{
				if(inPlayer->mNotPresent)
					return kFalse;

				inPlayer->mNotPresent = kTrue;
			}
			else
			{
				SendMsg(Msg_AboutToRemoveActivePlayer, Group_Network, inPlayer);
				mActivePlayers->Remove(iter.CurrentIndex());
			}

			return kTrue;
		}
	}

	return kFalse;
}

short
CTRINetwork::PingPlayer(SPlayer *inPlayer, Boolean inAsync)
{
	if(!HasConnection()) return -1;

	SendPacket(PacketPing::eID, inPlayer);
	if(!inAsync)
	{
		SyncWaitForPacket(PacketPingResponse::eID, kPingTimeout);
		return inPlayer->mPingTime;
	}

	return -1;
}

short
CTRINetwork::GetAvailablePlayerCount(void)
{
	return mAvailablePlayers->GetCount();
}

short
CTRINetwork::GetActivePlayerCount(void)
{
	return mActivePlayers->GetCount();
}

void
CTRINetwork::RandomizePlayerList(void)
{
//	if(mPlayersClosed) TRIThrow("Cannot randomize player list.  List is closed.", ERR_InternalError);
	if(mPlayersClosed) return;
	if(!IsHosting()) return;

	int count = mActivePlayers->GetUsed();
	for(int outer=0; outer < count*5; outer++)
	{
		int p1 = 1 + (rand() * count / RAND_MAX);
		int p2 = 1 + (rand() * count / RAND_MAX);

		if(p1 > count) p1 = count;
		if(p1 < 1) p1 = 1;
		if(p2 > count) p2 = count;
		if(p2 < 1) p2 = 1;

		SPlayer *player1 = mActivePlayers->Nth(p1);
		SPlayer *player2 = mActivePlayers->Nth(p2);

		mActivePlayers->Replace(p1, player2);
		mActivePlayers->Replace(p2, player1);
	}
}

void
CTRINetwork::ClosePlayerList(Boolean inAllowReplacement, Boolean inAllowAdditions)
{
	if(!HasConnection()) return;
	if(!IsHosting()) TRIThrow("Only host can close players", ERR_InternalError);

	mPlayersClosed=kTrue;
	mAllowReplacements = inAllowReplacement;
	mAllowAdditions = inAllowAdditions;

	// Tell everyone that the player list is closed
	SendPacket(PacketPlayerListStatus::eID, kAllPlayers);
	SendMsg(Msg_ClosePlayerList, Group_Network);

	// Remove all "Available players"
	SPlayerIter iter(mAvailablePlayers);
	while(!iter.AtEnd())
		RemovePlayer(iter.Next());

	LOGINDENT(1);
	LOG("Sending final player list");

	// Send finalize packets for each player and record local player index
	iter.SetList(mActivePlayers);
	while(!iter.AtEnd())
	{
		SPlayer *player = iter.Next();
		ElementIndex index = iter.CurrentIndex();
		if(player->mLocal)
			mLocalPlayerIndex = index;
		SendPacket(PacketFinalize::eID, kAllPlayers, player, NULL, &index);

#ifdef NETWORK_LOG
		char buf[256];
		sprintf(buf, "player: %s\tindex: %d", (const char*) *player->mName, index);
		LOG(buf);
#endif
	}

	LOGINDENT(-1);
}

void
CTRINetwork::OpenPlayerList(void)
{
	if(!IsHosting()) return;

	mPlayersClosed=kFalse;

	// Tell everyone that the player list is open
	SendPacket(PacketPlayerListStatus::eID, kAllPlayers);

	// remove all "NotPresent" players
	SPlayerIter iter(mActivePlayers);
	while(!iter.AtEnd())
	{
		SPlayer *player = iter.Next();
		if(player->mNotPresent)
			RemovePlayer(player);
	}

	// activate all "Inactive" players
	iter.SetList(mActivePlayers);
	iter.First();
	while(!iter.AtEnd())
	{
		SPlayer *player = iter.Next();
		if(player->mNotActive)
		{
			SendMsg(Msg_ActivatePlayer, Group_Network, player);
			player->mNotActive = kFalse;
		}
	}

	SendMsg(Msg_OpenPlayerList, Group_Network);
}

void
CTRINetwork::SendPacket(PacketID inID, const SPlayer *inPlayer, SPlayer *inPlayerRef,
						SPacket *inPacketRef, void *inOther, ENetPriority inPriority,
						ESendFlags inGuaranteed, short inCount)
{
	if(!HasConnection()) return;

	SPacket *packet = mNetworkManager->MakePacket(inID, inPlayerRef, inPacketRef, inOther);
	if(packet != NULL)
	{
		SendPacket(packet, inPlayer, inPriority, inGuaranteed, inCount);
		mNetworkManager->DeletePacket(packet);
	}
}

void
CTRINetwork::SendPacket(SPacket *inPacket, const SPlayer *inPlayer, ENetPriority inPriority, ESendFlags inGuaranteed, short inCount)
{
	if(!HasConnection()) return;

	if(inPacket != NULL)
	{
		inPacket->mFromID = mLocalPlayerID;
		for(short i=0; i<inCount; i++)
		{
			Boolean send = kTrue;
			if(inPlayer!=NULL && inPlayer->mLocal)
				send = GetDeliverLocalPackets();
			if(send)
				SendOnePacket(inPacket, inPlayer, inPriority, inGuaranteed);
		}
	}
}

void
CTRINetwork::ReceivePackets(Boolean inProcessMany)
{
	if(!HasConnection()) return;

	Boolean done = !inProcessMany;
	do {
		SPacket *packet = GetOnePacket();
		if(packet == NULL)
			done = kTrue;
		else
		{
			packet->mLocalData.mFrom = GetPlayer(packet->mFromID);
			if(!IsNetworkPacket(packet))
			{
				mNetworkManager->ProcessPacket(packet);
			}

			if(packet->mLocalData.mOwner == SPacket::eNetwork)
				delete packet;
			else
				mNetworkManager->DeletePacket(packet);
		}
	} while(!done);
}

void
CTRINetwork::RegisterNetworkManager(CNetworkManager *inManager)
{
	mNetworkManager = inManager;
	if(mNetworkManager == NULL)
		mNetworkManager = &gNetworkManager;
}

Boolean
CTRINetwork::SyncWaitForStatus(StatusType inStatus, short inTimeout)
{
	if(!HasConnection()) return kTrue;

	SPlayer *player = GetLocalPlayer();
	if(player == NULL) return kFalse;

	inTimeout *= kTimeToSeconds; // convert from seconds to getTime increments
	unsigned long startTime = getTime();
	unsigned long deltaTime = 0;

	while(player->mStatus!=inStatus && deltaTime<=inTimeout)
	{
		ReceivePackets();
		deltaTime = getTime() - startTime;
	}

	return (player->mStatus == inStatus);
}

Boolean
CTRINetwork::SyncWaitForPacket(PacketID inPacketID, short inTimeout)
{
	if(!HasConnection()) return kTrue;

	mWaitingForPacket = kTrue;
	mSyncWaitPacketType = inPacketID;

	inTimeout *= kTimeToSeconds; // convert from seconds to getTime increments
	unsigned long startTime = getTime();
	unsigned long deltaTime = 0;
	unsigned long timeOut = (unsigned short) inTimeout;

	while(mWaitingForPacket && deltaTime<=timeOut)
	{
		ReceivePackets();
		deltaTime = getTime() - startTime;
	}

	return !mWaitingForPacket;
}

Boolean
CTRINetwork::SyncWaitForPlayers(StatusType inStatus, PacketID inPacketID, short inTimeout)
{
	if(!HasConnection()) return kTrue;

	SPlayer *player = GetLocalPlayer();
	if(player != NULL)
		player->mStatus = inStatus;	// set the local player to inStatus

	inTimeout *= kTimeToSeconds; // convert from seconds to getTime increments
	long startTime = getTime();
	long deltaTime = 0;

	Boolean allReady = kTrue;
	do {
		ReceivePackets(kTrue);
		deltaTime = getTime() - startTime;

		allReady = kTrue;	// assume that everyone is ready

		SPlayerIter iter(mActivePlayers);
		while(!iter.AtEnd() && allReady)
		{
			SPlayer *player = iter.Next();
			if(!(player->mNotPresent || player->mNotActive))
				allReady = (player->mStatus == inStatus);
		}
	} while(!allReady && deltaTime<=inTimeout);

	return allReady;
}

Boolean
CTRINetwork::SyncSendWaitPacket(StatusType inStatus, PacketID inPacketID, short inDelay, short inTimeout)
{
	if(!HasConnection()) return kTrue;

	SPlayer *player = GetLocalPlayer();
	if(player != NULL)
		player->mStatus = inStatus;	// set the local player to inStatus

	inTimeout *= kTimeToSeconds; // convert from seconds to getTime increments
	inDelay *= kTimeToSeconds; // convert from seconds to getTime increments
	long startTime = getTime();
	long deltaTime = 0;
	long nextSendTime = 0;

	Boolean allReady = kTrue;
	do {
		if(deltaTime >= nextSendTime)
		{
			SendPacket(inPacketID, kAllPlayers);
			nextSendTime += inDelay;
		}

		ReceivePackets(kTrue);
		deltaTime = getTime() - startTime;

		allReady = kTrue;	// assume that everyone is ready

		SPlayerIter iter(mActivePlayers);
		while(!iter.AtEnd() && allReady)
		{
			SPlayer *player = iter.Next();
			if(!(player->mNotPresent || player->mNotActive))
				allReady = (player->mStatus == inStatus);
		}
	} while(!allReady && deltaTime<=inTimeout);

	return allReady;
}

Boolean
CTRINetwork::AddAvailablePlayer(SPlayer *inPlayer)
{
	if(!HasConnection()) return kFalse;

#ifdef _DEBUG
	if(IsPlayerAvailable(inPlayer) || IsPlayerActive(inPlayer))
		TRIThrow(String("Player previously added - ")+inPlayer->mName, ERR_InternalError);
#endif

	mAvailablePlayers->Insert(inPlayer);

	if(IsHosting())
	{
		SendPacket(PacketPlayerAvailable::eID, kAllPlayers, inPlayer);

		if(inPlayer->mLocal)
			inPlayer->mHost = kTrue;
		else
		{
			// Send player list to new player
			SPlayerIter iter(mAvailablePlayers);
			while(!iter.AtEnd())
			{
				SPlayer *player = iter.Next();
				if(player != inPlayer)
					SendPacket(PacketPlayerAvailable::eID, inPlayer, player);
			}

			PacketPlayerActive packet;
			iter.SetList(mActivePlayers);
			while(!iter.AtEnd())
			{
				SPlayer *player = iter.Next();
//				SendPacket(PacketPlayerActive::eID, inPlayer, player);

				if(!player->mNotPresent)
				{
					strcpy(packet.mName, *player->mName);
					packet.mID = player->mID;
					packet.mNotifyOnly = kTrue;
					SendPacket(&packet, inPlayer);
				}
			}

			// Send the host information to the new player
			SPlayer *host = GetHost();
			if(host != NULL)
				SendPacket(PacketChangeHost::eID, inPlayer, host);
		}
	}

	SendMsg(Msg_AddAvailPlayer, Group_Network, inPlayer);

	return kTrue;
}

Boolean
CTRINetwork::IsPlayerActive(SPlayer *inPlayer)
{
	if(!HasConnection()) return kFalse;

	SPlayerIter iter(mActivePlayers);
	while(!iter.AtEnd())
	{
		SPlayer *player = iter.Next();
		if(player == inPlayer)
			return kTrue;
	}

	return kFalse;
}

Boolean
CTRINetwork::IsPlayerAvailable(SPlayer *inPlayer)
{
	if(!HasConnection()) return kFalse;

	SPlayerIter iter(mAvailablePlayers);
	while(!iter.AtEnd())
	{
		SPlayer *player = iter.Next();
		if(player == inPlayer)
			return kTrue;
	}

	return kFalse;
}

void
CTRINetwork::RemoveAllPlayers(void)
{
	SendMsg(Msg_RemoveAllPlayers, Group_Network);

	SPlayerIter iter(mActivePlayers);
	while(!iter.AtEnd())
	{
		SPlayer *player = iter.Next();
		RemovePlayer(player);
	}

	iter.SetList(mAvailablePlayers);
	while(!iter.AtEnd())
	{
		SPlayer *player = iter.Next();
		RemovePlayer(player);
	}
}

void
CTRINetwork::BecomeHost(void)
{
	SPlayer *player = GetHost();
	if(player != NULL)
		player->mHost = kFalse;

	player = GetLocalPlayer();
	if(player == NULL) TRIThrow("Non-existing local player became host", ERR_InternalError);

	SendPacket(PacketChangeHost::eID, kAllPlayers, player);
	if(!GetDeliverLocalPackets())
	{	// packet was not delivered locally
		SetControl(eHosting);
		SendMsg(Msg_StartHosting, Group_Network);

		player->mHost = kTrue;
		SendMsg(Msg_HostChanged, Group_Network, player);
	}
}

Boolean
CTRINetwork::IsNetworkPacket(SPacket *inPacket)
{
	Boolean handled = kTrue;
	switch(inPacket->mID)
	{
		case PacketPlayerAvailable::eID:
			{
				PacketPlayerAvailable *packet = (PacketPlayerAvailable*) inPacket;
				if(GetPlayer(packet->mID) == NULL)
				{
					SPlayer *player = CreateNetPlayer(packet->mName, packet->mID);
				}
			}
			break;

		case PacketPlayerActive::eID:
			{
				PacketPlayerActive *packet = (PacketPlayerActive*) inPacket;
				SPlayer *player = GetPlayer(packet->mID);
				if(player == NULL)
				{
					player = CreateNetPlayer(packet->mName, packet->mID);
				}
				MakePlayerActive(player, packet->mNotifyOnly);
			}
			break;

		case PacketPlayerRemove::eID:
			{
			PacketPlayerRemove *packet = (PacketPlayerRemove*) inPacket;
			RemovePlayer(GetPlayer(packet->mID));
			}
			break;

		case PacketPing::eID:
			{
			SendPacket(PacketPingResponse::eID, inPacket->mLocalData.mFrom, NULL, inPacket);
			}
			break;

		case PacketPingResponse::eID:
			{
			if(inPacket->mLocalData.mFrom == NULL) break;
			PacketPingResponse *packet = (PacketPingResponse*) inPacket;
			unsigned long t = getTime();
			t -= packet->mSendTime;
			t *= 1000; // convert to milliseconds
			t /= kTimeToSeconds;
			unsigned long pingTime = t;
			inPacket->mLocalData.mFrom->mPingTime = pingTime;
			SendMsg(Msg_PlayerChanged, Group_Network, inPacket->mLocalData.mFrom);
			if(mWaitingForPacket && mSyncWaitPacketType==PacketPingResponse::eID)
				mWaitingForPacket = kFalse;
			else
				SendMsg(Msg_PingReceived, Group_Network, inPacket->mLocalData.mFrom);
			}
			break;

		case PacketNameChanged::eID:
			{
			PacketNameChanged *packet = (PacketNameChanged*) inPacket;
			SPlayer *player = GetPlayer(packet->mID);
			if(player != NULL)
			{
				*player->mName = packet->mNewName;
				SendMsg(Msg_PlayerChanged, Group_Network, player);
			}
			}
			break;

		case PacketPlayerListStatus::eID:
			mPlayersClosed = ((PacketPlayerListStatus*) inPacket)->mClosed;
			mAllowReplacements = ((PacketPlayerListStatus*) inPacket)->mAllowReplacements;
			mAllowAdditions = ((PacketPlayerListStatus*) inPacket)->mAllowAdditions;
			if(mPlayersClosed)
				SendMsg(Msg_ClosePlayerList, Group_Network);
			else
				SendMsg(Msg_OpenPlayerList, Group_Network);
			break;

		case PacketFinalize::eID:
			{
			PacketFinalize *packet = (PacketFinalize*) inPacket;

			LOGINDENT(1);
			LOG("Received finalize packet");

			// Find player
			SPlayerIter iter(mActivePlayers);
			while(!iter.AtEnd())
			{
				SPlayer *player = iter.Next();
				if(player->mID == packet->mID)
				{
					ElementIndex curIndex = iter.CurrentIndex();
					if(curIndex != packet->mIndex)
					{	// Place player in final position
						SPlayer *hold = (*mActivePlayers)[packet->mIndex];
						(*mActivePlayers)[packet->mIndex] = player;
						(*mActivePlayers)[curIndex] = hold;
					}

					if(player->mLocal)
						mLocalPlayerIndex = packet->mIndex;

#ifdef NETWORK_LOG
					char buf[256];
					sprintf(buf, "player: %s\tpacket index: %d\tcurIndex: %d\tlocal: %s",
						(const char*) *player->mName, packet->mIndex, curIndex, (player->mLocal?"Yes":"No"));
					LOG(buf);
#endif
					break;
				}
			}

			LOGINDENT(-1);
			}
			break;

		case PacketChangeHost::eID:
			{
				PacketChangeHost *packet = (PacketChangeHost*) inPacket;
				SPlayer *curHost = GetHost();
				SPlayer *player = GetPlayer(packet->mID);
				if(player == curHost) break;

				if(curHost != NULL)
				{
					curHost->mHost = kFalse;

					if(curHost->mLocal)
					{
						SetControl(eJoining);
						SendMsg(Msg_StopHosting, Group_Network);
					}
				}

				if(player != NULL)
				{
					SendMsg(Msg_HostChanged, Group_Network, player);
					player->mHost = kTrue;

					if(player->mLocal)
					{
						SetControl(eHosting);
						SendMsg(Msg_StartHosting, Group_Network);
					}
				}

			}
			break;

		default:
			return kFalse;
			break;
	}

	return handled;
}

////////////////////////////////////////////////////////////////
// Windows specific network support
#ifdef WIN
#include "stdafx.h"
#include "TRIForm.h"
#include "ResDefs.h"

#define INITGUID
#include <initguid.h> //modified - added this in for GUID_TCP to work
#include "DPLobby.h"
#undef INITGUID
#include "CGUID.h"

// Lobby support selector - should be set in ProjNet.h
//#define USELOBBY	// uses the lobby interface to start a connection

// Should ask app for GUID
extern GUID	gAppGUID;

// GUIDs for DirectPlay service providers
DEFINE_GUID(GUID_TCP, 
0x36E95EE0, 0x8577, 0x11cf, 0x96, 0xc, 0x0, 0x80, 0xc7, 0x53, 0x4e, 0x82);

DEFINE_GUID(GUID_MODEM, 
0x44eaa760, 0xcb68, 0x11cf, 0x9c, 0x4e, 0x0, 0xa0, 0xc9, 0x5, 0x42, 0x5e);

// These cannot be declared as members because DPLobby.h includes a windows
// header file and will stop the compile with #error.
LPDIRECTPLAY		gDP_1 = NULL;
LPDIRECTPLAY2A 		gDP_2 = NULL;
#ifdef USELOBBY
LPDIRECTPLAYLOBBYA	gDP_Lobby = NULL;
#endif

// These defines may be obsolete - check
//#define	MAX_DPLAY_SESSION_ENUM	32
//#define	MAX_DPLAY_SERVICE_ENUM	10
#define	SESSION_ENUM_TIMEOUT	1000

// type for dp objects
const unsigned long kDPType = '_DP_';

struct SDPSession : public SSession
{
	DPSESSIONDESC2	mDesc;

	SDPSession(const String& inName, const String& inPassword, short inMaxPlayers)
		: SSession(inName, inPassword, inMaxPlayers, kDPType)
		{mDesc.lpszSessionNameA=NULL;; mDesc.lpszPasswordA=NULL;};
	virtual ~SDPSession(void) {delete mDesc.lpszSessionNameA; delete mDesc.lpszPasswordA;};
};

struct SDPProtocol : public SProtocol
{
	GUID	mGUID;

	SDPProtocol(const String& inName) : SProtocol(inName, kDPType) {};
};

static int __stdcall AvailableNetServicesCallback(
		LPGUID lpGUID, LPSTR lpDriverDescription,
		DWORD dwMajorVersion, DWORD dwMinorVersion, LPVOID lpContext);

static int __stdcall AvailableNetSessionsCallback(
		LPCDPSESSIONDESC2 lpDPSGameDesc, LPDWORD lpdwTimeOut,
		DWORD dwFlags, LPVOID lpContext);

// default host setup dialog
class CDefHostDialog : public CTRIDialog
{
protected:
	SProtocolList	mProtocols;
	SProtocol		*mSelProtocol;
	String			mSessionName;

public:
	CDefHostDialog(CWnd* pParent = NULL);   // standard constructor

	inline SProtocol	*GetSelectedProtocol(void) {return mSelProtocol;};
	inline String&		GetSessionName(void) {return mSessionName;};

// Dialog Data
	//{{AFX_DATA(CDefHostDialog)
	enum { IDD = IDD_HOSTNETGAME };
	CListBox	m_Protocol;
	CEdit	m_SessionName;
	CButton	m_OK;
	CButton	m_Cancel;
	//}}AFX_DATA


// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CDefHostDialog)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

protected:
	// Generated message map functions
	//{{AFX_MSG(CDefHostDialog)
	virtual BOOL OnInitDialog();
	afx_msg void OnSelChangeProtocol();
	afx_msg void OnDblclkProtocol();
	afx_msg void OnChangeNewSessionName();
	virtual void OnCancel();
	virtual void OnOK();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};


// default join setup dialog
class CDefJoinDialog : public CTRIDialog
{
protected:
	SProtocolList	mProtocols;
	SSessionList	mSessions;
	BOOL			m_DoAuto;
	SProtocol		*mSelProtocol;
	SSession		*mSelSession;

public:
	CDefJoinDialog(CWnd* pParent = NULL);   // standard constructor

	inline SSession		*GetSelectedSession(void){return mSelSession;};
	inline SProtocol	*GetSelectedProtocol(void) {return mSelProtocol;};

	// Dialog Data
	//{{AFX_DATA(CDefJoinDialog)
	enum { IDD = IDD_JOINNETGAME };
	CButton	m_AutoSelect;
	CListBox	m_SessionList;
	CListBox	m_Protocol;
	CButton	m_OK;
	CButton	m_Cancel;
	//}}AFX_DATA

	// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CDefJoinDialog)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

protected:
	void	FindSessions(void);

	// Generated message map functions
	//{{AFX_MSG(CDefJoinDialog)
	virtual BOOL OnInitDialog();
	afx_msg void OnSelChangeProtocol();
	afx_msg void OnSelChangeSession();
	virtual void OnCancel();
	virtual void OnOK();
	afx_msg void OnDblclkSession();
	afx_msg void OnTimer(UINT nIDEvent);
	afx_msg void OnAutoselect();
	afx_msg void OnDblclkProtocol();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};


// CDefJoinDialog Implementation
CDefHostDialog::CDefHostDialog(CWnd* pParent /*=NULL*/)
	: CTRIDialog(CDefHostDialog::IDD)
{
	//{{AFX_DATA_INIT(CDefHostDialog)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT

	m_bitmapName[0] = 0;
	mSelProtocol = NULL;
}

void CDefHostDialog::DoDataExchange(CDataExchange* pDX)
{
	CTRIDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CDefHostDialog)
	DDX_Control(pDX, IDC_PROTOCOL, m_Protocol);
	DDX_Control(pDX, IDC_NEWSESSIONNAME, m_SessionName);
	DDX_Control(pDX, IDOK, m_OK);
	DDX_Control(pDX, IDCANCEL, m_Cancel);
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CDefHostDialog, CTRIDialog)
	//{{AFX_MSG_MAP(CDefHostDialog)
	ON_LBN_SELCHANGE(IDC_PROTOCOL, OnSelChangeProtocol)
	ON_LBN_DBLCLK(IDC_PROTOCOL, OnDblclkProtocol)
	ON_EN_CHANGE(IDC_NEWSESSIONNAME, OnChangeNewSessionName)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

BOOL CDefHostDialog::OnInitDialog() 
{
	CTRIDialog::OnInitDialog();
	
//	m_OK.Init("OKBtn.bmp", &m_palette);
//	m_Cancel.Init("CanclBtn.bmp", &m_palette);
	
#ifdef IDS_SESSIONDEFAULT
	CString sesName;
	sesName.LoadString(IDS_SESSIONDEFAULT);
	m_SessionName.SetWindowText(sesName);
#endif

	if(gNetwork->GetProtocolList(mProtocols))
	{
		SProtocolIter iter(&mProtocols);
		while(!iter.AtEnd())
		{
			SProtocol *p = iter.Next();
			m_Protocol.AddString(*p->mName);
		}

		OnSelChangeProtocol();
	}

	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void CDefHostDialog::OnSelChangeProtocol() 
{
	m_OK.EnableWindow(m_Protocol.GetCurSel() >= 0 && m_SessionName.GetWindowTextLength()>0);
}

void CDefHostDialog::OnDblclkProtocol() 
{
	if ((m_Protocol.GetCurSel() >= 0)
		&& (m_SessionName.GetWindowTextLength() > 0))
	{
		OnOK();
	}	
}

void CDefHostDialog::OnChangeNewSessionName() 
{
	OnSelChangeProtocol();
}

void CDefHostDialog::OnOK() 
{
	mSelProtocol = mProtocols.Remove(m_Protocol.GetCurSel()+1);
	CString name;
	m_SessionName.GetWindowText(name);
	mSessionName = name.GetBuffer(1);

	gNetwork->ClearProtocolList(mProtocols);
	CTRIDialog::OnOK();
}

void CDefHostDialog::OnCancel() 
{
	gNetwork->ClearProtocolList(mProtocols);
	CTRIDialog::OnCancel();
}


// CDefJoinDialog Implementation
CDefJoinDialog::CDefJoinDialog(CWnd* pParent /*=NULL*/)
	: CTRIDialog(CDefJoinDialog::IDD)
{
	//{{AFX_DATA_INIT(CDefJoinDialog)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT

	m_bitmapName[0] = 0;
	mSelProtocol = NULL;
	mSelSession = NULL;
}

void CDefJoinDialog::DoDataExchange(CDataExchange* pDX)
{
	CTRIDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CDefJoinDialog)
	DDX_Control(pDX, IDC_AUTOSELECT, m_AutoSelect);
	DDX_Control(pDX, IDC_SESSION, m_SessionList);
	DDX_Control(pDX, IDC_PROTOCOL, m_Protocol);
	DDX_Control(pDX, IDOK, m_OK);
	DDX_Control(pDX, IDCANCEL, m_Cancel);
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CDefJoinDialog, CTRIDialog)
	//{{AFX_MSG_MAP(CDefJoinDialog)
	ON_LBN_SELCHANGE(IDC_PROTOCOL, OnSelChangeProtocol)
	ON_LBN_SELCHANGE(IDC_SESSION, OnSelChangeSession)
	ON_LBN_DBLCLK(IDC_SESSION, OnDblclkSession)
	ON_WM_TIMER()
	ON_BN_CLICKED(IDC_AUTOSELECT, OnAutoselect)
	ON_LBN_DBLCLK(IDC_PROTOCOL, OnDblclkProtocol)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

BOOL CDefJoinDialog::OnInitDialog() 
{
	CTRIDialog::OnInitDialog();
	
//	m_OK.Init("OKBtn.bmp", &m_palette);
//	m_Cancel.Init("CanclBtn.bmp", &m_palette);
	
	m_DoAuto = TRUE;
	m_AutoSelect.SetCheck(m_DoAuto);

	if(gNetwork->GetProtocolList(mProtocols))
	{
		SProtocolIter iter(&mProtocols);
		while(!iter.AtEnd())
		{
			SProtocol *p = iter.Next();
			m_Protocol.AddString(*p->mName);
		}

		OnSelChangeProtocol();
	}

	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void CDefJoinDialog::FindSessions(void)
{
	KillTimer(m_timerRef[0]);
	m_SessionList.ResetContent();

	if(gNetwork->GetSessionList(mProtocols[m_Protocol.GetCurSel()+1], mSessions))
	{
		SSessionIter iter(&mSessions);
		while(!iter.AtEnd())
		{
			SSession *ses = (SSession*) iter.Next();
			m_SessionList.AddString(*ses->mName);
		}

		// we want to check for new net sessions every few seconds -- setup a timer...
		m_timerRef[0] = SetTimer('sess', 1000, NULL);
	}
	else
	{
		m_OK.EnableWindow(FALSE);
	}
}

void CDefJoinDialog::OnSelChangeProtocol() 
{
	if(!m_DoAuto) return;

	KillTimer(m_timerRef[0]);
	gNetwork->Disconnect();

	m_OK.EnableWindow(FALSE);
	if(m_Protocol.GetCurSel() >= 0)
	{
		FindSessions();
	}
}

void CDefJoinDialog::OnDblclkProtocol() 
{
	if(m_DoAuto) return;

	KillTimer(m_timerRef[0]);
	gNetwork->Disconnect();

	m_OK.EnableWindow(FALSE);
	if(m_Protocol.GetCurSel() >= 0)
	{
		FindSessions();
	}
}

void CDefJoinDialog::OnSelChangeSession() 
{
	m_OK.EnableWindow(m_SessionList.GetCurSel() >= 0);
}

void CDefJoinDialog::OnDblclkSession() 
{
	OnOK();
}

void CDefJoinDialog::OnTimer(UINT nIDEvent) 
{
	if (nIDEvent == 'sess')
	{
		gNetwork->ClearSessionList(mSessions);
		if(gNetwork->GetSessionList(mProtocols[m_Protocol.GetCurSel()+1], mSessions))
		{
			int listSel = m_SessionList.GetCurSel();
			m_SessionList.ResetContent();

			SSessionIter iter(&mSessions);
			while(!iter.AtEnd())
			{
				SSession *ses = (SSession*) iter.Next();
				m_SessionList.AddString(*ses->mName);
			}		

			// default to first available session
			if (mSessions.GetCount() > 0)
			{
				if (mSessions.GetCount() > listSel)
				{
					m_SessionList.SetCurSel(listSel);
					m_OK.EnableWindow(TRUE);
				} else
				{
					m_SessionList.SetCurSel(0);
					m_OK.EnableWindow(FALSE);
				}
			} else
			{				
				m_OK.EnableWindow(FALSE);
			}
		}
		else
		{
			KillTimer(m_timerRef[0]);
			m_SessionList.ResetContent();
			m_OK.EnableWindow(FALSE);
		}
	}
	
	CTRIDialog::OnTimer(nIDEvent);
}

void CDefJoinDialog::OnAutoselect() 
{
	m_DoAuto = m_AutoSelect.GetCheck();
}

void CDefJoinDialog::OnOK() 
{
	KillTimer(m_timerRef[0]);

	mSelProtocol = mProtocols.Remove(m_Protocol.GetCurSel()+1);
	mSelSession = (SSession*) mSessions.Remove(m_SessionList.GetCurSel()+1);

	gNetwork->ClearProtocolList(mProtocols);
	gNetwork->ClearSessionList(mSessions);
	CTRIDialog::OnOK();
}

void CDefJoinDialog::OnCancel() 
{
	KillTimer(m_timerRef[0]);
	gNetwork->ClearProtocolList(mProtocols);
	gNetwork->ClearSessionList(mSessions);
	gNetwork->Disconnect();
	CTRIDialog::OnCancel();
}


// CTRIDPlay Implementation
CTRIDPlay::CTRIDPlay(void)
{
	SetStatus(eNotConnected);

	// Enumeration control
	mEnumState = eStopped;
	mSessionList = NULL;
	mProtocolList = NULL;
}

CTRIDPlay::~CTRIDPlay(void)
{
	Disconnect();
}

Boolean
CTRIDPlay::StartAutoConnect(SPlayer *inPlayer)
{
#ifdef USELOBBY
	if(HasConnection() || !CanAutoConnect()) return kFalse;

	LOGINDENT(1);
	LOG("CTRIDPlay::StartAutoConnect");

	Boolean result = FALSE;

	if(gDP_Lobby == NULL)
	{
		LOG(" - Start Commandline");
		result = DoCommandLineLobby(inPlayer);
	}
	else
	{
		LOG(" - Do True lobby");
		result = DoTrueLobby(inPlayer);
	}

	LOGINDENT(-1);
	return result;

#else	// !USELOBBY
	// Don't use the lobby, just connect using the default method
	return CTRINetwork::StartAutoConnect(inPlayer);
#endif	// USELOBBY
}

Boolean
CTRIDPlay::DoTrueLobby(SPlayer *inPlayer)
{
#ifdef USELOBBY

// Use the DPlay lobby object as it was meant to be ???
#define CHECK(test) if(test) {success=kFalse; goto cleanup;}

	LOGINDENT(1);
	LOG("CTRIDPlay::DoTrueLobby");

	Boolean success = kTrue;
	HRESULT result = DP_OK;
	LPDPLCONNECTION connSettings = NULL;

	LOG(" - Get Address size");
	unsigned long addrSize = 0;
	result = gDP_Lobby->GetConnectionSettings(0, NULL, &addrSize);
	CHECK(result != DPERR_BUFFERTOOSMALL);

	connSettings = (LPDPLCONNECTION) new char[addrSize];
	CHECK(connSettings == NULL);

	LOG(" - Get Connection settings");
	result = gDP_Lobby->GetConnectionSettings(0, connSettings, &addrSize);
	CHECK(result != DP_OK);

	connSettings->lpSessionDesc->dwFlags |= DPSESSION_KEEPALIVE;
	if(mAllowHostMigration)
		connSettings->lpSessionDesc->dwFlags |= DPSESSION_MIGRATEHOST;

	LOG(" - Set Connection settings");
	connSettings->lpSessionDesc->dwMaxPlayers = 8;
	result = gDP_Lobby->SetConnectionSettings(0, 0, connSettings);
	CHECK(result != DP_OK);

	LOG(" - Connect");
	result = gDP_Lobby->Connect(0, &gDP_2, NULL);
	CHECK(result != DP_OK);

	LOG(" - Set status");
	SetControl((connSettings->dwFlags & DPLCONNECTION_CREATESESSION) ? eHosting : eJoining);
	SetStatus(eConnected);

	LOG(" - Use local player");
	UseLocalPlayer(inPlayer);

	SendMsg(Msg_Connect, Group_Network);

	LOG(" - Successful");

cleanup:
	if(!success)
	{
#ifdef NETWORK_LOG
		char temp[256];
		sprintf(temp, " - Error code:  %X", result);
		LOG(temp);
#endif
	}

	if(!success && gDP_2!=NULL)
	{
		LOG(" - gDP_2->Release");
		gDP_2->Release();
		gDP_2 = NULL;
	}

	if(gDP_Lobby != NULL)
	{
		LOG(" - gDP_Lobby->Release");
		gDP_Lobby->Release();
		gDP_Lobby = NULL;
	}

	LOG(" - Complete");
	LOGINDENT(-1);

	delete connSettings;
	return success;

#undef CHECK

#else
	return kFalse;
#endif
}

Boolean
CTRIDPlay::DoCommandLineLobby(SPlayer *inPlayer)
{
#ifdef USELOBBY

// This uses the command-line parameters and DPlay's lobby object
// to create or join a session
#define CHECK(test) if(test) {success=kFalse; goto cleanup;}

	String defName("Lobby test session");
	String defPassword;

#ifdef IDS_SESSIONDEFAULT
	CString sesName;
	sesName.LoadString(IDS_SESSIONDEFAULT);
	defName = sesName.GetBuffer(1);
#endif

	defName % *inPlayer->mName;

	Boolean success = kTrue;
	HRESULT result = DP_OK;

	SDPProtocol *protocol = (SDPProtocol*) GetAutoConnectProtocol(mAutoConn->mProtocolSelector);
	CHECK(protocol == NULL);

	result = DirectPlayLobbyCreate(NULL, &gDP_Lobby, NULL, NULL, 0);
	CHECK(result != DP_OK);

	DPSESSIONDESC2 sessionDesc;
	sessionDesc.dwSize = sizeof(sessionDesc);
	sessionDesc.dwFlags = 0;
	sessionDesc.dwFlags = DPSESSION_KEEPALIVE;
	if(mAllowHostMigration) sessionDesc.dwFlags |= DPSESSION_MIGRATEHOST;
	sessionDesc.guidInstance = GUID_NULL;
	sessionDesc.guidApplication = gAppGUID;
	sessionDesc.dwMaxPlayers = 8;
	sessionDesc.dwCurrentPlayers = 0;
#ifdef DBCS
	sessionDesc.lpszSessionName = (short*) (const short*) defName;
	if(defPassword.Length() == 0)
		sessionDesc.lpszPassword = NULL;
	else
		sessionDesc.lpszPassword = (short*) (const short*) defPassword;
#else
	sessionDesc.lpszSessionNameA = (char*) (const char*) defName;
	if(defPassword.Length() == 0)
		sessionDesc.lpszPassword = NULL;
	else
		sessionDesc.lpszPasswordA = (char*) (const char*) defPassword;
#endif
	sessionDesc.dwReserved1 = 0;
	sessionDesc.dwReserved2 = 0;
	sessionDesc.dwUser1 = 0;
	sessionDesc.dwUser2 = 0;
	sessionDesc.dwUser3 = 0;
	sessionDesc.dwUser4 = 0;

	DPNAME playerName;
	playerName.dwSize = sizeof(playerName);
	playerName.dwFlags = 0;
	playerName.lpszShortNameA = (char*) (const char*) *inPlayer->mName;
	playerName.lpszLongNameA = (char*) (const char*) *inPlayer->mName;

	DPLCONNECTION connInfo;
	connInfo.dwSize = sizeof(connInfo);
	connInfo.lpSessionDesc = &sessionDesc;
	connInfo.lpPlayerName = &playerName;
	connInfo.guidSP = protocol->mGUID;

	if(mAutoConn->mHost)
	{
		connInfo.dwFlags = DPLCONNECTION_CREATESESSION;
		connInfo.lpAddress = NULL;
		connInfo.dwAddressSize = 0;
	}
	else
	{
		unsigned long	addrSize = 0;
		unsigned long	addrLength = mAutoConn->mAddress->Length()+1;
		char	addrText[256];
		mAutoConn->mAddress->CopyBuffer(addrText, 256);

		result = gDP_Lobby->CreateAddress(protocol->mGUID, DPAID_INet,
			addrText, addrLength, NULL, &addrSize);
		CHECK(result != DPERR_BUFFERTOOSMALL);

		LPVOID address = new char[addrSize];
		CHECK(address == NULL);

		result = gDP_Lobby->CreateAddress(protocol->mGUID, DPAID_INet,
			addrText, addrLength, address, &addrSize);
		CHECK(result != DP_OK);
		
		connInfo.dwFlags = DPLCONNECTION_JOINSESSION;
		connInfo.lpAddress = address;
		connInfo.dwAddressSize = addrSize;
	}

	result = gDP_Lobby->SetConnectionSettings(0, 0, &connInfo);
	CHECK(result != DP_OK);

	result = gDP_Lobby->Connect(0, &gDP_2, NULL);
	CHECK(result != DP_OK);

	SetControl(mAutoConn->mHost ? eHosting : eJoining);
	SetStatus(eConnected);
	UseLocalPlayer(inPlayer);
	SendMsg(Msg_Connect, Group_Network);

cleanup:
	if(!success && gDP_2!=NULL)
	{
		gDP_2->Release();
		gDP_2 = NULL;
	}

	if(gDP_Lobby != NULL)
	{
		gDP_Lobby->Release();
		gDP_Lobby = NULL;
	}

	delete mAutoConn;
	mAutoConn = NULL;

	delete connInfo.lpAddress;
	delete protocol;

	return success;

#undef CHECK

#else
	return kFalse;
#endif
}

Boolean
CTRIDPlay::CanAutoConnect(void)
{
	Boolean result = kFalse;

#ifdef USELOBBY
	if(HasConnection())
		return kFalse;

	// The application will call this method once but we don't
	// want to keep killing the lobby object so we will save it
	// between calls.
	if(gDP_Lobby != NULL) return kTrue;

	LOGINDENT(1);
	LOG("CTRIDPlay::CanAutoConnect");

	LOG(" - create lobby");
	HRESULT hresult = DirectPlayLobbyCreate(NULL, &gDP_Lobby, NULL, NULL, 0);
	if(gDP_Lobby != NULL)
	{
		LOG(" - get connection settings");
		unsigned long addrSize = 0;
		hresult = gDP_Lobby->GetConnectionSettings(0, NULL, &addrSize);
		if(hresult == DPERR_NOTLOBBIED)
		{
			LOG(" - not lobbied");
			gDP_Lobby->Release();
			gDP_Lobby = NULL;
		}
		else
		{
			LOG(" - is lobbied");
			result = kTrue;
		}
	}

	// If command-line info has been posted we may be able to autoconnect
	// without having been lobbied
#endif

	if(!result)
	{
		LOG(" - trying command line auto connect");
		result = CTRINetwork::CanAutoConnect();
	}

	LOGINDENT(-1);
	return result;
}

Boolean
CTRIDPlay::StartHost(SPlayer *inPlayer)
{
	CDefHostDialog dialog;
	if(dialog.DoModal() == IDOK)
	{
		SProtocol *protocol = dialog.GetSelectedProtocol();
		SDPSession *session = new SDPSession(dialog.GetSessionName(), "", 8);
		if(OpenHostSession(protocol, session))
		{
			UseLocalPlayer(inPlayer);
			return kTrue;
		}
	}

	return kFalse;
}

Boolean
CTRIDPlay::StartHost(const String& inPlayerName)
{
	CDefHostDialog dialog;
	if(dialog.DoModal() == IDOK)
	{
		SProtocol *protocol = dialog.GetSelectedProtocol();
		SDPSession *session = new SDPSession(dialog.GetSessionName(), "", 8);
		if(OpenHostSession(protocol, session))
		{
			CreateLocalPlayer(inPlayerName);
			return kTrue;
		}
	}

	return kFalse;
}

Boolean
CTRIDPlay::StartJoin(SPlayer *inPlayer)
{
	CDefJoinDialog dialog;
	if(dialog.DoModal() == IDOK)
	{
		SProtocol *protocol = dialog.GetSelectedProtocol();
		SSession *session = dialog.GetSelectedSession();

		if(OpenJoinSession(protocol, session))
		{
			UseLocalPlayer(inPlayer);
			return kTrue;
		}
	}

	return kFalse;
}

Boolean
CTRIDPlay::StartJoin(const String& inPlayerName)
{
	CDefJoinDialog dialog;
	if(dialog.DoModal() == IDOK)
	{
		SProtocol *protocol = dialog.GetSelectedProtocol();
		SSession *session = dialog.GetSelectedSession();

		if(OpenJoinSession(protocol, session))
		{
			CreateLocalPlayer(inPlayerName);
			return kTrue;
		}
	}

	return kFalse;
}

Boolean
CTRIDPlay::GetProtocolList(SProtocolList& outList)
{
	mProtocolList = &outList;
	StartEnum();
	HRESULT result = DirectPlayEnumerate(AvailableNetServicesCallback, NULL);
	StopEnum();
	mProtocolList = NULL;

	return (result == DP_OK);
}

Boolean
CTRIDPlay::GetSessionList(SProtocol*inProtocol, SSessionList& outList)
{
	if(inProtocol != GetCurrentProtocol())
		Disconnect();

	if(!OpenNetworkConnection(inProtocol))
		return kFalse;

	// Find out what DirectPlay services are available to us...
	DPSESSIONDESC2	sessionDesc;
	memset(&sessionDesc, 0, sizeof(sessionDesc));
	sessionDesc.dwSize = sizeof(sessionDesc);
	sessionDesc.guidInstance = GUID_NULL;
	sessionDesc.guidApplication = gAppGUID;

	mSessionList = &outList;
	StartEnum();
	HRESULT result = gDP_2->EnumSessions(&sessionDesc, SESSION_ENUM_TIMEOUT,
		AvailableNetSessionsCallback, NULL, DPENUMSESSIONS_AVAILABLE);
	StopEnum();
	mSessionList = NULL;

	return (result == DP_OK);
}

SSession*
CTRIDPlay::CreateSessionObject(const String& inName, const String& inPassword, short inMaxPlayers)
{
	return new SDPSession(inName, inPassword, inMaxPlayers);
}

Boolean
CTRIDPlay::OpenNetworkConnection(SProtocol *inProtocol)
{
	if(inProtocol == NULL) return kFalse;

	if(gDP_2 != NULL) return kTrue;	// Already intialized

	SDPProtocol *dpProtocol = (SDPProtocol*) inProtocol;
	HRESULT result = DirectPlayCreate(&dpProtocol->mGUID, &gDP_1, NULL);
	
	if(result == DP_OK)
	{
		result = gDP_1->QueryInterface(IID_IDirectPlay2A, (LPVOID*) &gDP_2);
		if(result == DP_OK)
		{
			mCurProtocol = inProtocol;

			// KLUDGE - slow down packet transmission for modem connections
			// with more than two players (i.e. TCP/IP)
			if(memcmp(&GUID_TCP, &dpProtocol->mGUID, sizeof(GUID)) == 0)
				SetBestPacketRate(4);
			else
				SetBestPacketRate(8);
		}
	}

	return result == DP_OK;
}

Boolean
CTRIDPlay::OpenJoinSession(SProtocol *inProtocol, SSession *inSession)
{
	if(inSession == NULL) return kFalse;

	if(!OpenNetworkConnection(inProtocol))
		return kFalse;

	SDPSession *dpSession = (SDPSession*) inSession;
	dpSession->mDesc.dwFlags = DPOPEN_OPENSESSION | DPSESSION_KEEPALIVE;
	if(mAllowHostMigration)
		dpSession->mDesc.dwFlags |= DPSESSION_MIGRATEHOST;
	HRESULT result = gDP_2->Open(&dpSession->mDesc, DPOPEN_JOIN);

	if(result == DP_OK)
	{
		SetStatus(eConnected);
		SetControl(eJoining);
		SetCurrentProtocol(inProtocol);
		SetCurrentSession(inSession);
		SendMsg(Msg_Connect, Group_Network);
	}
	return result==DP_OK;
}

Boolean
CTRIDPlay::OpenHostSession(SProtocol *inProtocol, SSession *inSession)
{
	if(inSession == NULL) return kFalse;

	if(!OpenNetworkConnection(inProtocol))
		return kFalse;

	SDPSession *dpSession = (SDPSession*) inSession;

	dpSession->mDesc.lpszSessionNameA = new char[dpSession->mName->Length() + 1];
	strcpy(dpSession->mDesc.lpszSessionNameA, *dpSession->mName);

	dpSession->mDesc.lpszPasswordA = new char[dpSession->mPassword->Length() + 1];
	strcpy(dpSession->mDesc.lpszPasswordA, *dpSession->mPassword);

	dpSession->mDesc.dwSize = sizeof(dpSession->mDesc);
	dpSession->mDesc.guidInstance = GUID_NULL;
	dpSession->mDesc.guidApplication = gAppGUID;
	dpSession->mDesc.dwMaxPlayers = inSession->mMaxPlayers;
	dpSession->mDesc.dwCurrentPlayers = 0;
	dpSession->mDesc.dwFlags = DPOPEN_CREATESESSION | DPSESSION_KEEPALIVE;
	if(mAllowHostMigration)
		dpSession->mDesc.dwFlags |= DPSESSION_MIGRATEHOST;

	HRESULT result = gDP_2->Open(&dpSession->mDesc, DPOPEN_CREATE);
	if(result == DP_OK)
	{
		SetStatus(eConnected);
		SetControl(eHosting);
		SetCurrentProtocol(inProtocol);
		SetCurrentSession(inSession);
		SendMsg(Msg_Connect, Group_Network);
	}
	else
		Disconnect();

	return result==DP_OK;
}

Boolean
CTRIDPlay::Disconnect(void)
{
	Boolean result = CTRINetwork::Disconnect();

	if (gDP_2 != NULL)
	{
		LOGINDENT(1);
		LOG("CTRIDPlay::Disconnect");

		// close the session
		gDP_2->Close();
		gDP_2->Release();			// free the direct play object
		gDP_2 = NULL;

		if(gDP_1 != NULL)
		{
			gDP_1->Close();
			gDP_1->Release();
			gDP_1 = NULL;
		}

		LOGINDENT(-1);
	}

	return result;
}

Boolean
CTRIDPlay::UseLocalPlayer(SPlayer *inPlayer)
{
	LOGINDENT(1);
	LOG("CTRIDPlay::UseLocalPlayer");

	Boolean valid = HasConnection() && inPlayer!=NULL;
	if(!valid)
	{
		LOGINDENT(-1);
		return kFalse;
	}

	LOG(" - GetLocalPlayer");
	Boolean success = FALSE;

	SPlayer *oldPlayer = GetLocalPlayer();
	if(oldPlayer == NULL)
	{	// Local player is being used for the first time
		LOG(" - New local player");

		char *buffer = inPlayer->mName->CopyBuffer();
		DPNAME name;
		name.dwSize = sizeof(name);
		name.dwFlags = 0;
		name.lpszShortNameA = buffer;
		name.lpszLongNameA = buffer;
		HRESULT result = gDP_2->CreatePlayer(&mLocalPlayerID, &name, NULL, NULL, 0, 0);
		delete buffer;

		if(result == DP_OK)
		{
			inPlayer->mID = mLocalPlayerID;
			inPlayer->mLocal = kTrue;
			if(AddAvailablePlayer(inPlayer))
				MakePlayerActive(inPlayer, kFalse);

			LOG(" - Success");

			success = kTrue;
		}
	}
	else
	{	// Local player is being changed
		LOG(" - Change local player");
		success = CTRINetwork::UseLocalPlayer(inPlayer);
	}

	LOG(" - Failed");
	LOGINDENT(-1);
	return success;
}

void
CTRIDPlay::ChangedLocalPlayer(void)
{
	// TODO - tell DPlay that the players name changed
	CTRINetwork::ChangedLocalPlayer();
}

Boolean
CTRIDPlay::RemovePlayer(SPlayer *inPlayer)
{
	if(inPlayer == NULL) return kFalse;
	if(!HasConnection()) return kFalse;

	Boolean isLocal = inPlayer->mLocal;
	PlayerID id = inPlayer->mID;
	if(CTRINetwork::RemovePlayer(inPlayer))
	{
		if(isLocal)
		{
			if(id == mLocalPlayerID)
				mLocalPlayerID = 0;
			if(gDP_2 != NULL)
				gDP_2->DestroyPlayer(id);
		}

		return kTrue;
	}

	return kFalse;
}

SProtocol*
CTRIDPlay::GetAutoConnectProtocol(unsigned long inProtocol)
{
	SProtocolList list;
	GetProtocolList(list);
	
	const GUID *autoGUID = &GUID_TCP;

	SDPProtocol *theProtocol = NULL;
	SProtocolIter iter(&list);
	while(!iter.AtEnd())
	{
		SDPProtocol *protocol = (SDPProtocol*) iter.Next();
		if(memcmp(&protocol->mGUID, autoGUID, sizeof(GUID)) == 0)
		{
			list.Remove(iter.CurrentIndex());
			theProtocol = protocol;
			break;
		}
	}

	ClearProtocolList(list);

	return theProtocol;
}

SSession*
CTRIDPlay::GetAutoConnectSession(SProtocol *inProtocol, Boolean inHost)
{
	if(inProtocol == NULL) return NULL;

	String defName("AutoConnect test session");
	String defPassword;

#ifdef IDS_SESSIONDEFAULT
	CString sesName;
	sesName.LoadString(IDS_SESSIONDEFAULT);
	defName = sesName.GetBuffer(1);
#else
		// TODO - where to get the default session name and password
#endif

	if(inHost)
	{
		return new SDPSession(defName, defPassword, 8);
	}
	else
	{
		SSessionList list;
		GetSessionList(inProtocol, list);

		SSessionIter iter(&list);
		while(!iter.AtEnd())
		{
			SSession *session = (SSession*) iter.Next();
			if(*session->mName == defName)
			{
				list.Remove(iter.CurrentIndex());
				ClearSessionList(list);
				return session;
			}
		}

		return NULL;
	}
}

SPlayer*
CTRIDPlay::CreateNetPlayer(const String& inName, PlayerID inID)
{
	SPlayer *player = mNetworkManager->CreatePlayer(inName, inID, kDefaultType, SPlayer::eNetwork);
	if(player == NULL) TRIThrow("Could not create net player", ERR_OutOfMemory);

	if(!AddAvailablePlayer(player))
	{
		delete player;
		player = NULL;
	}
	else
	{
		PingPlayer(player);
	}

	return player;
}

unsigned long
CTRIDPlay::CountPendingPackets(void)
{
	if(!HasConnection()) return 0;

	unsigned long count;
	gDP_2->GetMessageCount(mLocalPlayerID, &count);
	return count;
}

void
CTRIDPlay::SendOnePacket(SPacket *inPacket, const SPlayer *inPlayer, ENetPriority inPriority, ESendFlags inGuaranteed)
{
	DWORD sendFlags;
	switch (inGuaranteed)
	{
		case eGuaranteed:
			sendFlags = DPSEND_GUARANTEE;
			break;
		case eNonGuaranteed:
		default:
			sendFlags = DPSEND_TRYONCE;
			break;
	}

	switch(inPriority)
	{
		case eNetHigh:
			sendFlags |= DPSEND_HIGHPRIORITY;
			break;
		case eNetLow:
		case eNetNormal:
		default:
			break;
	}

	if (gDP_2 == NULL) return;
	if (mLocalPlayerID == 0) return;

	PlayerID sendTo = (inPlayer == NULL) ? DPID_ALLPLAYERS : inPlayer->mID;

	// the local data does not need to be sent so we can remove it from the packet
	const long skipBytes = sizeof(SPacket::SLocalData);

	HRESULT result = gDP_2->Send(mLocalPlayerID, sendTo, sendFlags,
		((char*)inPacket) + skipBytes, inPacket->mLocalData.mSize - skipBytes);

#ifdef _DEBUG
//	if(result != DP_OK)
//	{
//		String errorCode(result, kTrue);
//		TRIThrow(String("DPSend failed - ") + errorCode, ERR_InternalError);
//	}
#endif

#ifdef NETWORK_LOG
	{
	char temp[256];
	char ID[5];
	ID[0] = (inPacket->mID & 0xFF000000) >> 24;
	ID[1] = (inPacket->mID & 0x00FF0000) >> 16;
	ID[2] = (inPacket->mID & 0x0000FF00) >> 8;
	ID[3] = (inPacket->mID & 0x000000FF) >> 0;
	ID[4]=0;
	sprintf(temp, "Sent packet (\'%s\' : %X)", ID, inPacket->mID);
	LOG(temp);

	if(result != DP_OK)
	{
		sprintf(temp, "CTRIDPlay::SendOnePacket failed:  %X", result);
		LOG(temp);
	}
	}
#endif
}

SPacket*
CTRIDPlay::GetOnePacket(void)
{
	if (gDP_2 == NULL) return NULL;

	DWORD count;
	gDP_2->GetMessageCount(mLocalPlayerID, &count);

	if (count == 0) return NULL;

	// Get messages that are to this player
	const short kMaxPacketSize = 4096;
	DWORD	size = kMaxPacketSize;
	DPID	from;
	char	staticpacket[kMaxPacketSize];
	char	*packet = staticpacket;
	int returnValue = gDP_2->Receive(&from, &mLocalPlayerID, DPRECEIVE_TOPLAYER,
		(void *) packet, &size);  

	// If the static buffer is too small, create a temporary larger one
	Boolean deleteBuffer = kFalse;
	if(returnValue == DPERR_BUFFERTOOSMALL)
	{
		deleteBuffer = kTrue;
		size += sizeof(SPacket::SLocalData);
		packet = new char[size];
		returnValue = gDP_2->Receive(&from, &mLocalPlayerID, DPRECEIVE_TOPLAYER,
			(void *) packet, &size);  
	}

	// If no messages then don't cry about it after all
	if (returnValue == DPERR_NOMESSAGES) 
		return NULL;

	// Any other error, bail!
	if(returnValue != DP_OK)
	{
		if(deleteBuffer)
			delete packet;
		return NULL;
	}

	// messages from me to me are messages I sent to the server and it
	// is echoing them back
	if (from == mLocalPlayerID) 
	{
		if(deleteBuffer)
			delete packet;
		return NULL;
	}

	LOGINDENT(1);
	LOG("CTRIDPlay::GetOnePacket");

	// Setup default return value
	SPacket *p = NULL;

	// if message is from user zero then it is a Direct Play system message
	if (from == 0) 
	{	// process Direct Play system messages
		DPMSG_GENERIC *pGeneric = (DPMSG_GENERIC *) packet;
		switch (pGeneric->dwType) 
		{
			case DPSYS_HOST:
				LOG(" - Become Host");
				BecomeHost();
				break;

			case DPSYS_SESSIONLOST:
				LOG(" - Session Lost");
				Disconnect();
				break;

			case DPSYS_DESTROYPLAYERORGROUP:
				{
					LOG(" - Destroy Player");
					DPMSG_DESTROYPLAYERORGROUP *pMsg = (DPMSG_DESTROYPLAYERORGROUP *) packet;
					RemovePlayer(GetPlayer(pMsg->dpId));
				}
				break;

			case DPSYS_CREATEPLAYERORGROUP:
				{
					LOG(" - Create Player");
					DPMSG_CREATEPLAYERORGROUP *pMsg = (DPMSG_CREATEPLAYERORGROUP*) packet;
					if (pMsg->dwPlayerType == DPPLAYERTYPE_PLAYER)
					{
						CreateNetPlayer(pMsg->dpnName.lpszLongNameA, pMsg->dpId);
					}
				}
				break;

			// unsupported system messages 
			case DPSYS_ADDPLAYERTOGROUP:
			case DPSYS_DELETEPLAYERFROMGRP:
			default:
				LOG(" - Unsupported System Message");
				break;
		}
//
//		return NULL;
	}
	else
	{	// Packet is from a player
		LOG(" - Received from another player");

		const long extraSize = sizeof(SPacket::SLocalData);

		p = (SPacket*) new char[extraSize + size];
//		SPacket *p = (SPacket*) new char[extraSize + size];

		memcpy(((char*)p)+extraSize, packet, size);
		p->mLocalData.mSize = extraSize+size;
		p->mLocalData.mFrom = NULL;
		p->mLocalData.mOwner = SPacket::eNetwork;

#ifdef NETWORK_LOG
	{
	char temp[256];
	char ID[5];
	ID[0] = (p->mID & 0xFF000000) >> 24;
	ID[1] = (p->mID & 0x00FF0000) >> 16;
	ID[2] = (p->mID & 0x0000FF00) >> 8;
	ID[3] = (p->mID & 0x000000FF) >> 0;
	ID[4]=0;
	sprintf(temp, "Received packet (\'%s\' : %X)", ID, p->mID);
	LOG(temp);
	}
#endif

//		return p;
	}

	LOGINDENT(-1);

	if(deleteBuffer)
		delete packet;

	return p;
//	return NULL;
}

short
CTRIDPlay::AddSessionToList(SSession *inSession)
{
	if(mSessionList != NULL)
	{
		mSessionList->Insert(inSession);
		return mSessionList->GetCount();
	}

	return 0;
}

short
CTRIDPlay::AddProtocolToList(SProtocol *inProtocol)
{
	if(mProtocolList != NULL)
	{
		mProtocolList->Insert(inProtocol);
		return mProtocolList->GetCount();
	}

	return 0;
}

int __stdcall AvailableNetServicesCallback(
	LPGUID lpGUID,
	LPSTR lpDriverDescription,
	DWORD dwMajorVersion,
	DWORD dwMinorVersion,
	LPVOID lpContext)
{
	SDPProtocol *protocol = new SDPProtocol(lpDriverDescription);
	if(protocol == NULL) TRIThrow("Could not allocate protocol", ERR_OutOfMemory);
	protocol->mGUID = *lpGUID;
	short count = ((CTRIDPlay*)gNetwork)->AddProtocolToList(protocol);

	if (((CTRIDPlay*)gNetwork)->GetEnumState() == CTRIDPlay::eWaitingToStop)
		return(kFalse);
	else
		return kTrue;
//		return (count < MAX_DPLAY_SERVICE_ENUM);
}

int __stdcall AvailableNetSessionsCallback(
	LPCDPSESSIONDESC2 lpDPSGameDesc,
	LPDWORD lpdwTimeOut,
	DWORD dwFlags,
	LPVOID lpContext)
{
	if (dwFlags == DPESC_TIMEDOUT)	return(FALSE);

	short count = 0;
	if (lpDPSGameDesc != NULL)
	{
		SDPSession *ses = new SDPSession(lpDPSGameDesc->lpszSessionNameA,
							"", lpDPSGameDesc->dwMaxPlayers);

		ses->mDesc.dwSize = lpDPSGameDesc->dwSize;
		ses->mDesc.dwFlags = lpDPSGameDesc->dwFlags;
		ses->mDesc.guidInstance = lpDPSGameDesc->guidInstance;
		ses->mDesc.guidApplication = lpDPSGameDesc->guidApplication;
		ses->mDesc.dwMaxPlayers = lpDPSGameDesc->dwMaxPlayers;
		ses->mDesc.dwCurrentPlayers = lpDPSGameDesc->dwCurrentPlayers;

		if(lpDPSGameDesc->lpszSessionNameA != NULL)
		{
			ses->mDesc.lpszSessionNameA = new char[ strlen(lpDPSGameDesc->lpszSessionNameA) + 1 ];
			strcpy(ses->mDesc.lpszSessionNameA, lpDPSGameDesc->lpszSessionNameA);
		}
		else
			ses->mDesc.lpszSessionNameA = NULL;

		if(lpDPSGameDesc->lpszPasswordA != NULL)
		{
			ses->mDesc.lpszPasswordA = new char[ strlen(lpDPSGameDesc->lpszPasswordA) + 1 ];
			strcpy(ses->mDesc.lpszPasswordA, lpDPSGameDesc->lpszPasswordA);
		}
		else
			ses->mDesc.lpszPasswordA = NULL;

		ses->mDesc.dwReserved1 = lpDPSGameDesc->dwReserved1;
		ses->mDesc.dwReserved2 = lpDPSGameDesc->dwReserved2;
		ses->mDesc.dwUser1 = lpDPSGameDesc->dwUser1;
		ses->mDesc.dwUser2 = lpDPSGameDesc->dwUser2;
		ses->mDesc.dwUser3 = lpDPSGameDesc->dwUser3;
		ses->mDesc.dwUser4 = lpDPSGameDesc->dwUser4;

		count = ((CTRIDPlay*)gNetwork)->AddSessionToList(ses);
	}

	if (((CTRIDPlay*)gNetwork)->GetEnumState() == CTRIDPlay::eWaitingToStop)
		return kFalse;
	else
		return kTrue;
//		return ((count < MAX_DPLAY_SESSION_ENUM) && (lpDPSGameDesc != NULL));
}

#endif // WIN


////////////////////////////////////////////////////////////////
// Macintosh specific network support
#ifdef MAC
#endif // MAC


////////////////////////////////////////////////////////////////
// PC specific network support
#ifdef PC
#endif //PC

